/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#ifndef DB_H
#define DB_H 1


#define LIBFLUX_BUILD 1


#ifdef LIBFLUX_BUILD
# include "types.h"
# include "misc.h"
# include "tt.h"
# include "bmem.h"
# include "cache.h"
#else
# include <flux/types.h>
# include <flux/misc.h>
# include <flux/tt.h>
# include <flux/bmem.h>
# include <flux/cache.h>
#endif


#define DB_VERSION 0
#define DB_FILE_META  "meta"
#define DB_FILE_TREE  "tree"
#define DB_FILE_INDEX "index"


typedef enum
{
  DB_TREE_FIRST_CHILD, DB_TREE_LAST_CHILD, DB_TREE_BEFORE, DB_TREE_AFTER
}
DB_TREE_RELATIVE;



extern char *db_file_data[];


struct pt_node
{
  u32 byte;   /* Index of byte to compare. */
  u32 id;     /* Data ID - 4 category bits, 28 index bits. */
  u32 left;   /* 3 MSbits: Index of bit to compare. */
  u32 right;  /* 3 MSbits: Unused. */
};

struct data_head
{
  u32 size;
  u32 ref;
  u32 flags;
};

struct tree_node
{
  u32 parent;
  u32 child_first;
  u32 child_last;
  u32 next;
  u32 prev;
  u32 data;
};


typedef struct _db
{
  char *path;  /* Path to base directory */
  TT *root;    /* In-memory root node */

  CACHE *pt_cache;
  CACHE *tt_cache;
  u32 nextid[16];  /* Next free ID for each of our categories */
  int datafd[16];  /* Current filehandle for each data file */
  int treefd;      /* Current filehandle for the tree structure */
  int indexfd;     /* Current filehandle for the data index */
}
DB;


/* --- Allocation --- */

DB *db_new(char *path);
DB *db_open(char *path);
void db_close(DB *db);
void db_del(DB *db);

/* --- Miscellaneous --- */

TT *db_get_root(DB *db);

/* --- Files --- */

void db_file_meta_create(DB *db);
int db_file_meta_check(DB *db);

void db_file_index_create(DB *db);
int db_file_index_open(DB *db);

int db_file_data_open(DB *db, int cat);

int db_file_tree_open(DB *db);

/* --- Data --- */

#define db_data_get_id(db, cat) ((db)->nextid[cat] | (cat) << 28)
#define db_data_incr_id(db, cat) (db)->nextid[cat]++

struct data_head *db_data_head_read(int fd, byte cat);
void db_data_head_write(int fd, byte cat, struct data_head *dh);

int db_data_fd(DB *db, u32 id);
int db_data_fd_with_pos_and_size(DB *db, u32 id, u32 *size);

int db_data_bit(DB *db, u32 id, u32 byte, unsigned char bit);
int db_data_last_diff_bit_mem(DB *db, u32 id, byte *data, u32 len, u32 *byt, byte *bit);
int db_data_compare_mem(DB *db, u32 id, byte *data, u32 len);

/* --- Data indexing --- */

#define db_index_open(db) (db)->indexfd = db_file_index_open(db)
#define db_index_close(db) close((db)->indexfd), (db)->indexfd = -1

void db_index_read(DB *db, int fd, u32 block, struct pt_node **pt);
void db_index_write(DB *db, int fd, u32 block, struct pt_node *pt);

u32 db_index_new(int fd);
u32 db_index_insert_mem(DB *db, u32 id, unsigned char *data, unsigned int len);
u32 db_index_find(DB *db, unsigned char *data, unsigned int len);

/* --- Tree structure --- */

#define db_tree_open(db) (db)->treefd = db_file_tree_open(db)
#define db_tree_close(db) close((db)->treefd), (db)->treefd = -1


#endif /* DB_H */
