/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#ifndef _SMTP_H
#define _SMTP_H 1

#ifndef LIBFLUX_BUILD
#  include <flux/sock.h>
#  include <flux/nvtp.h>
#  include <flux/tt.h>
#else
#  include "sock.h"
#  include "nvtp.h"
#  include "tt.h"
#endif /* LIBFLUX_BUILD */


typedef enum
{
  /* Minimum supported */
  SMTP_CMD_HELLO, SMTP_CMD_MAIL, SMTP_CMD_RECIPIENT, SMTP_CMD_DATA,
  SMTP_CMD_NOOP, SMTP_CMD_RESET, SMTP_CMD_QUIT,

  /* Send to TTY commands - not supported */
  SMTP_CMD_SEND, SMTP_CMD_SEND_OR_MAIL, SMTP_CMD_SEND_AND_MAIL,

  /* Extra server commands - not supported yet */
  SMTP_CMD_VERIFY, SMTP_CMD_EXPAND, SMTP_CMD_HELP, SMTP_CMD_TURN,

  /* Special cases */
  
  SMTP_CMD_BLANK, SMTP_CMD_UNKNOWN
}
SMTP_CMD_CODE;


typedef enum
{
  /* Status/help - usually human-readable stuff, which can be ignored */
  SMTP_REPLY_STATUS_OR_HELP      = 211,
  SMTP_REPLY_HELP                = 214,

  /* Service information */
  SMTP_REPLY_SERVICE_READY       = 220,
  SMTP_REPLY_CLOSING_CHANNEL     = 221,
  SMTP_REPLY_SERVICE_UNAVAILABLE = 421,
  
  /* Syntax errors */
  SMTP_REPLY_SYNTAX_ERR_CMD      = 500,
  SMTP_REPLY_SYNTAX_ERR_ARG      = 501,
  SMTP_REPLY_UNIMPLEMENTED_CMD   = 502,
  SMTP_REPLY_UNIMPLEMENTED_ARG   = 504,
  SMTP_REPLY_BAD_SEQUENCE        = 503,
  
  /* OK replies */
  SMTP_REPLY_OK                  = 250,
  SMTP_REPLY_WILL_FORWARD        = 251,
  SMTP_REPLY_START_MAIL_DATA     = 354,
  
  /* Delivery problems */
  SMTP_REPLY_MAILBOX_BUSY        = 450,
  SMTP_REPLY_MAILBOX_UNAVAILABLE = 550,
  SMTP_REPLY_ABORTED             = 451,
  SMTP_REPLY_USER_NOT_LOCAL      = 551,
  SMTP_REPLY_NO_STORAGE          = 452,
  SMTP_REPLY_EXCEEDED_STORAGE    = 552,
  SMTP_REPLY_MAILBOX_INVALID     = 553,
  SMTP_REPLY_TRANSACTION_FAILED  = 554
}
SMTP_REPLY_CODE;


/* Our transaction role */

enum smtp_role
{
  SMTP_ROLE_SERVER = 0, SMTP_ROLE_CLIENT
};


/* State of incoming queue - next item expected */

enum smtp_state_client
{
  SMTP_STATE_IDLE = 0,          /* Client, not waiting for a reply (ready) */
  SMTP_STATE_HELLO,             /* Client, waiting for HELO ack */
  SMTP_STATE_MAIL,              /* Client, waiting for MAIL ack */
  SMTP_STATE_RECIPIENT,         /* Client, waiting for RCPT ack */
  SMTP_STATE_DATA_BEGIN,        /* Client, waiting for DATA ack */
  SMTP_STATE_DATA_END,          /* Client, waiting for "." ack */
  SMTP_STATE_NOOP,              /* Client, waiting for NOOP ack */
  SMTP_STATE_RESET,             /* Client, waiting for RSET ack */
  SMTP_STATE_QUIT               /* Client, waiting for QUIT ack */
};

enum smtp_state_server
{
  SMTP_STATE_COMMAND_INIT = 0,  /* Server, initial cmds: helo, mail, noop, rset, quit */
  SMTP_STATE_COMMAND_MAIL,      /* Server, mail cmds: rcpt, data, rset */
  SMTP_STATE_DATA_HEAD,         /* Server, data head: header or blank line */
  SMTP_STATE_DATA_BODY          /* Server, data body: data line or "." */
};


typedef struct _smtp
{
  NVTP *nvtp;

  struct _smtp *parent;

  TT *queue_in, *queue_out;

  TT *root_in, *root_out;
  TT *node_in, *node_out;

  u16 state_in : 4;
  u16 state_out : 4;
  u16 role : 1;
}
SMTP;


/* --- */

SMTP *_smtp_alloc();
SMTP *smtp_new();
SMTP *smtp_pipe_new();
SMTP *smtp_new_with_port(int port);
SMTP *smtp_accepting_new(int port);
SMTP *smtp_new_with_sock(SOCK *s);
SMTP *smtp_new_with_nvtp(NVTP *nvtp);
void smtp_del(SMTP *smtp);

#define smtp_get_parent(smtp) (smtp)->parent

#define smtp_queue_out_len(smtp) tt_count_children(((SMTP *) smtp)->queue_out)
#define smtp_queue_in_len(smtp) tt_count_children(((SMTP *) smtp)->queue_in)

#define smtp_buf_out_size(smtp) nvtp_buf_out_size(((SMTP *) smtp)->nvtp)
#define smtp_buf_in_size(smtp) nvtp_buf_in_size(((SMTP *) smtp)->nvtp)

#define smtp_get_sock(smtp) nvtp_get_sock(((SMTP *) smtp)->nvtp)
#define smtp_get_nvtp(smtp) ((SMTP *) smtp)->nvtp
#define smtp_get_fd_read(smtp) nvtp_get_fd_read(smtp_get_nvtp(smtp))
#define smtp_get_fd_write(smtp) nvtp_get_fd_write(smtp_get_nvtp(smtp))

#define smtp_is_connecting(smtp) nvtp_is_connecting(smtp_get_nvtp(smtp))
#define smtp_is_connected(smtp) nvtp_is_connected(smtp_get_nvtp(smtp))
#define smtp_is_accepting(smtp) nvtp_is_accepting(smtp_get_nvtp(smtp))
#define smtp_is_active(smtp) (smtp_is_connected(smtp) | smtp_is_accepting(smtp))

SMTP *smtp_accept(SMTP *smtp);

void smtp_push(SMTP *smtp);
void smtp_pull(SMTP *smtp);

void smtp_enqueue(SMTP *smtp, TT *tt);
TT *smtp_dequeue(SMTP *smtp);

void smtp_reply(SMTP *smtp, SMTP_REPLY_CODE code, char *string);
int smtp_parse_reply(SMTP *smtp, TT *tt);

void smtp_cmd_hello(SMTP *smtp, char *host);
void smtp_cmd_mail(SMTP *smtp, char *from);
void smtp_cmd_recipient(SMTP *smtp, char *recipient);
void smtp_cmd_data(SMTP *smtp);
void smtp_cmd_noop(SMTP *smtp);
void smtp_cmd_reset(SMTP *smtp);
void smtp_cmd_quit(SMTP *smtp);

int smtp_parse_cmd(TT *tt, TT **arg);
TT *smtp_parse_arg_address(TT *cmd);

#endif /* _SMTP_H */
